import inspect
from typing import Optional, TypeVar, Union, List, Tuple, Callable, Dict, Any
from ._handlers import Handler, MiddlewareHandler
from ._router import Router
from ..client import Bot
from ..types import Update, Message, CallbackQuery
from ..stats.storage import BaseStorage, MemoryStorage
from ..stats import StatsManager
from ..utils.callback_data import CallbackData
from ._handlers import FallbackHandler, ErrorHandler

U = TypeVar("U", bound=Union[Update, Message, CallbackQuery])


class Dispatcher:
    def __init__(self, bot: Bot, storage: Optional[BaseStorage] = None):
        self.bot = bot
        self.routers: List[Router] = []
        self.storage = storage or MemoryStorage()
        self.fallback = FallbackHandler()
        self.error = ErrorHandler()

    def include_router(self, router: "Router") -> None:
        self.routers.append(router)

    async def process(self, update: Update) -> None:
        try:
            handler, middlewares = await self._find_handler(update=update)
            if not handler:
                if self.fallback.handler:
                    kwargs = await self._build_handler_kwargs(
                        self.fallback.handler, update, {}
                    )
                    await self.fallback.handler(**kwargs)
                return

            final_handler = await self._build_final_handler(handler.callback, update)
            wrapped_handler = self._wrap_middlewares(
                middlewares.middlewares, final_handler
            )

            await wrapped_handler(update, {})

        except Exception as e:
            await self._handle_error(e, update)

    async def _handle_error(self, error: Exception, update: Update) -> None:
        for exception_type, handler in self.error.handlers:
            if exception_type is not None and isinstance(error, exception_type):
                await handler(error, update)
                return

        for exception_type, handler in self.error.handlers:
            if exception_type is None:
                await handler(error, update)
                return

        raise error

    def _wrap_middlewares(
        self, middlewares: List[Callable], final_handler: Callable
    ) -> Callable:
        handler = final_handler
        for middleware in reversed(middlewares):
            handler = self._create_middleware_wrapper(middleware, handler)
        return handler

    def _create_middleware_wrapper(
        self, middleware: Callable, next_handler: Callable
    ) -> Callable:
        async def wrapper(update: Update, data: Dict[str, Any]) -> Any:
            return await middleware(next_handler, update, data)

        return wrapper

    async def _build_final_handler(self, handler: Callable, update: Update) -> Callable:
        async def final_handler(update: Update, data: Dict[str, Any]) -> Any:
            kwargs = await self._build_handler_kwargs(handler, update, data)
            return await handler(**kwargs)

        return final_handler

    async def _find_handler(
        self, update: Update
    ) -> Optional[Tuple[Handler, MiddlewareHandler]]:
        stats = await self.storage.get_stats(update.origin.from_user.chatid)
        for router in self.routers:
            handler = await router.matches_update(update=update, stats=stats)
            if handler:
                return handler, router.middleware
        return None, None

    async def _build_handler_kwargs(
        self, handler: Callable, update: Update, middleware_data: Dict[str, Any]
    ) -> Dict[str, Any]:
        sig = inspect.signature(handler)
        kwargs = {}
        origin = update.origin

        for name, value in middleware_data.items():
            if name in sig.parameters:
                kwargs[name] = value

        common_params = {
            "update": update,
            "stats": StatsManager(
                key=int(origin.from_user.chatid), storage=self.storage
            ),
            "bot": self.bot,
            "message": update.message,
            "callback_query": update.callback_query,
        }

        for name, value in common_params.items():
            if name in sig.parameters and value is not None:
                kwargs[name] = value

        if (
            update.callback_query
            and "callback_data" in sig.parameters
            and inspect.isclass(sig.parameters["callback_data"].annotation)
            and issubclass(sig.parameters["callback_data"].annotation, CallbackData)
        ):
            kwargs["callback_data"] = sig.parameters["callback_data"].annotation.unpack(
                update.callback_query.data
            )

        for param in sig.parameters:
            if param not in kwargs:
                if hasattr(update, param):
                    kwargs[param] = getattr(update, param)
                elif param in update.data:
                    kwargs[param] = update.data[param]

        return kwargs



from typing import Optional, Union
from eiogram.types import Message, InlineKeyboardMarkup
from ._base import MethodBase


class EditMessage(MethodBase):
    async def execute(
        self,
        text: str,
        chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None,
    ) -> Message:
        data = {
            "text": text,
            "parse_mode": "HTML",
        }

        if reply_markup:
            data["reply_markup"] = reply_markup.dict()

        if inline_message_id:
            data["inline_message_id"] = inline_message_id
        else:
            data.update({"chat_id": chat_id, "message_id": message_id})

        response = await self._make_request("POST", "editMessageText", data)
        return Message(**response["result"])



from ._answer_callback import AnswerCallbackQuery
from ._delete_messages import DeleteMessages
from ._edit_message import EditMessage
from ._get_me import GetMe
from ._restrict_user import RestrictUser
from ._send_message import SendMessage
from ._send_photo import SendPhoto
from ._set_my_commands import SetMyCommands
from ._delete_webhook import DeleteWebhook
from ._set_webhook import SetWebhook
from ._pin_message import PinMessage
from ._get_chat_member import GetChatMember

__all__ = [
    "AnswerCallbackQuery",
    "DeleteMessages",
    "EditMessage",
    "GetMe",
    "RestrictUser",
    "SendMessage",
    "SendPhoto",
    "SetMyCommands",
    "DeleteWebhook",
    "SetWebhook",
    "PinMessage",
    "GetChatMember",
]



from typing import List
from ...middleware import BaseMiddleware


class MiddlewareHandler:
    def __init__(self):
        self.middlewares: List[BaseMiddleware] = []

    def register(self, middleware: BaseMiddleware) -> BaseMiddleware:
        self.middlewares.append(middleware)
        self.middlewares.sort(key=lambda m: m.priority, reverse=True)
        return middleware



from ._base import Filter
from ._chat_type import IsSuperGroup, IsChannel, IsForum, IsGroup, IsPrivate
from ._command import Command, StartCommand
from ._data import Data
from ._photo import Photo
from ._regax import Regex
from ._text import Text
from ._stats import StatsFilter

__all__ = [
    "BaseTextFilter",
    "Filter",
    "IsSuperGroup",
    "IsChannel",
    "IsForum",
    "IsGroup",
    "IsPrivate",
    "Command",
    "StartCommand",
    "Data",
    "Photo",
    "Regex",
    "Text",
    "StatsFilter",
]



from typing import List, Union
from ._base import MethodBase


class DeleteMessages(MethodBase):
    async def execute(self, chat_id: Union[int, str], message_ids: List[int]) -> bool:
        data = {"chat_id": chat_id, "message_ids": message_ids}
        response = await self._make_request("POST", "deleteMessages", data)
        return response.get("result", False)



from typing import Optional, Union
from pydantic import BaseModel
from ..client import Bot


class User(BaseModel):
    id: int
    is_bot: bool
    first_name: str
    last_name: Optional[str] = None
    username: Optional[str] = None
    bot: Optional[Bot] = None

    class Config:
        arbitrary_types_allowed = True

    @property
    def full_name(self) -> str:
        return (
            f"{self.first_name} {self.last_name}" if self.last_name else self.first_name
        )

    @property
    def chatid(self) -> int:
        return self.id

    @property
    def mention(self) -> Optional[str]:
        return f"@{self.username}" if self.username else None

    def __str__(self) -> str:
        return (
            f"User(id={self.id}, "
            f"name={self.full_name}, "
            f"username={self.username or 'N/A'})"
        )

    async def is_join(
        self,
        chat_id: Union[str, int],
    ) -> bool:
        from ._chat import ChatMemberStatus

        try:
            status = await self.bot.get_chat_member(chat_id=chat_id, user_id=self.id)
            if status and status in [
                ChatMemberStatus.ADMINISTRATOR,
                ChatMemberStatus.CREATOR,
                ChatMemberStatus.MEMBER,
                ChatMemberStatus.RESTRICTED,
            ]:
                return True
        except Exception:
            return False

    async def is_admin(
        self,
        chat_id: Union[str, int],
    ) -> bool:
        from ._chat import ChatMemberStatus

        try:
            status = await self.bot.get_chat_member(chat_id=chat_id, user_id=self.id)
            if status and status in [
                ChatMemberStatus.ADMINISTRATOR,
                ChatMemberStatus.CREATOR,
            ]:
                return True
        except Exception:
            return False



from typing import List, Union, Optional, Tuple, Dict, Any
from ..types._inline_keyboard import InlineKeyboardMarkup, InlineKeyboardButton


class InlineKeyboardBuilder:
    def __init__(self):
        self._keyboard: List[List[InlineKeyboardButton]] = []

    def add(
        self,
        text: str,
        callback_data: Optional[str] = None,
        url: Optional[str] = None,
        web_app: Optional[str] = None,
        copy_text: Optional[str] = None,
    ) -> "InlineKeyboardBuilder":
        if not any([callback_data, url, web_app, copy_text]):
            raise ValueError("At least one button action must be specified")

        button = InlineKeyboardButton(
            text=text,
            callback_data=callback_data,
            web_app=web_app,
            url=url,
            copy_text=copy_text,
        )

        if not self._keyboard:
            self._keyboard.append([])
        self._keyboard[-1].append(button)
        return self

    def row(
        self, *buttons: InlineKeyboardButton, size: Union[int, Tuple[int, ...]] = None
    ) -> "InlineKeyboardBuilder":
        if not buttons:
            return self

        if size is None:
            self._keyboard.append(list(buttons))
        elif isinstance(size, int):
            for i in range(0, len(buttons), size):
                self._keyboard.append(list(buttons[i : i + size]))
        else:
            buttons = list(buttons)
            for s in size:
                if s > 0 and buttons:
                    self._keyboard.append(buttons[:s])
                    buttons = buttons[s:]
            if buttons:
                self._keyboard.append(buttons)
        return self

    def adjust(self, *sizes: Union[int, Tuple[int, ...]]) -> "InlineKeyboardBuilder":
        buttons = [btn for row in self._keyboard for btn in row]
        self._keyboard = []

        if not sizes:
            if buttons:
                self._keyboard = [buttons]
            return self

        for size in sizes:
            if isinstance(size, int):
                if size > 0 and buttons:
                    self._keyboard.append(buttons[:size])
                    buttons = buttons[size:]
            else:
                for s in size:
                    if s > 0 and buttons:
                        self._keyboard.append(buttons[:s])
                        buttons = buttons[s:]

        if buttons:
            self._keyboard.append(buttons)
        return self

    def as_markup(self) -> InlineKeyboardMarkup:
        clean_keyboard = [row for row in self._keyboard if row]
        return InlineKeyboardMarkup(inline_keyboard=clean_keyboard)

    def export(self) -> List[List[Dict[str, Any]]]:
        return [
            [
                {
                    k: v
                    for k, v in {
                        "text": btn.text,
                        "callback_data": btn.callback_data,
                        "url": btn.url,
                        "web_app": {"url": btn.web_app} if btn.web_app else None,
                        "copy_text": {"text": btn.copy_text} if btn.copy_text else None,
                    }.items()
                    if v is not None
                }
                for btn in row
                if any(
                    [
                        btn.callback_data,
                        btn.url,
                        btn.web_app,
                        btn.copy_text,
                    ]
                )
            ]
            for row in self._keyboard
            if row
        ]

    def __len__(self) -> int:
        return sum(len(row) for row in self._keyboard)

    @property
    def keyboard(self) -> List[List[InlineKeyboardButton]]:
        return [row for row in self._keyboard if row]



from ._error import ErrorHandler
from ._callback_query import CallbackQueryHandler
from ._message import MessageHandler
from ._base import Handler, BaseHandler, FilterFunc, HandlerFunc
from ._middleware import MiddlewareHandler
from ._fallback import FallbackHandler

__all__ = [
    "FallbackHandler",
    "ErrorHandler",
    "CallbackQueryHandler",
    "MessageHandler",
    "MiddlewareHandler",
    "Handler",
    "BaseHandler",
    "FilterFunc",
    "HandlerFunc",
]



from typing import Optional, List
from pydantic import BaseModel


class InlineKeyboardButton(BaseModel):
    text: str
    callback_data: Optional[str] = None
    url: Optional[str] = None
    web_app: Optional[str] = None
    copy_text: Optional[str] = None

    def dict(self) -> dict:
        result = {"text": self.text}
        if self.callback_data is not None:
            result["callback_data"] = self.callback_data
        if self.url is not None:
            result["url"] = self.url
        if self.web_app is not None:
            result["web_app"] = {"url": self.web_app}
        if self.copy_text is not None:
            result["copy_text"] = {"text": self.copy_text}
        return result


class InlineKeyboardMarkup(BaseModel):
    inline_keyboard: List[List[InlineKeyboardButton]]

    def dict(self) -> dict:
        return {
            "inline_keyboard": [
                [button.dict() for button in row] for row in self.inline_keyboard
            ]
        }



from typing import Optional, Union, Any, Dict
from .storage import BaseStorage, MemoryStorage
from ._stats import State


class StatsManager:
    def __init__(self, key: Union[int, str], storage: BaseStorage = MemoryStorage()):
        self.storage = storage
        self.key = key

    async def set_state(self, state: State) -> None:
        await self.storage.set_stats(self.key, stats=state.name)

    async def get_state(self) -> Optional[State]:
        return await self.storage.get_stats(self.key)

    async def upsert_data(self, **data: Any) -> None:
        return await self.storage.upsert_data(self.key, **data)

    async def get_data(self) -> Dict[str, Any]:
        return await self.storage.get_data(self.key)

    async def clear_all(self) -> None:
        await self.storage.clear_all(self.key)

    async def clear_stats(self) -> None:
        await self.storage.clear_stats(self.key)

    async def clear_data(self) -> None:
        await self.storage.clear_data(self.key)



from ._base import Filter


class _ChatTypeFilter(Filter):
    """Base class for chat type filters"""

    def __init__(self, chat_type: str):
        super().__init__(
            lambda msg: (
                hasattr(msg, "chat")
                and hasattr(msg.chat, "type")
                and msg.chat.type == chat_type
            )
        )


class IsPrivate(_ChatTypeFilter):
    def __init__(self):
        super().__init__("private")


class IsGroup(_ChatTypeFilter):
    def __init__(self):
        super().__init__("group")


class IsSuperGroup(_ChatTypeFilter):
    def __init__(self):
        super().__init__("supergroup")


class IsChannel(_ChatTypeFilter):
    def __init__(self):
        super().__init__("channel")


class IsForum(_ChatTypeFilter):
    def __init__(self):
        super().__init__("forum")



from typing import Optional
from ...types import User
from ._base import MethodBase


class GetMe(MethodBase):
    async def execute(self) -> Optional[User]:
        response = await self._make_request("GET", "getMe")
        return User(**response["result"])



from ._base import Filter


class Photo(Filter):
    def __init__(self):
        super().__init__(
            lambda msg: (
                hasattr(msg, "photo") and msg.photo is not None and bool(msg.photo)
            )
        )



from typing import Callable, Awaitable, TypeVar

from ._base import BaseHandler, FilterFunc
from ...types import Message

MessageT = TypeVar("MessageT", bound=Message)
MessageHandlerFunc = Callable[[MessageT], Awaitable[None]]


class MessageHandler(BaseHandler):
    def __init__(self):
        super().__init__(update_type="message")

    def __call__(
        self, *filters: FilterFunc, priority: int = 0
    ) -> Callable[[MessageHandlerFunc], MessageHandlerFunc]:
        return super().__call__(*filters, priority=priority)



from typing import Optional, Union
from ...types import Message, InlineKeyboardMarkup
from ._base import MethodBase


class SendMessage(MethodBase):
    async def execute(
        self,
        chat_id: Union[int, str],
        text: str,
        reply_markup: Optional[InlineKeyboardMarkup] = None,
        reply_to_message_id: Optional[int] = None,
    ) -> Optional[Message]:
        data = {
            "chat_id": chat_id,
            "text": text,
            "parse_mode": "HTML",
        }

        if reply_markup:
            data["reply_markup"] = reply_markup.dict()

        if reply_to_message_id:
            data["reply_to_message_id"] = reply_to_message_id

        response = await self._make_request("POST", "sendMessage", data)
        return Message(**response["result"])



from typing import Optional
from ._base import Filter
from ..stats import State


class StatsFilter(Filter):
    def __init__(self, state: Optional[State]):
        super().__init__(lambda stats: (stats is not None and stats == state.name))



from typing import Any, Callable, Dict, Awaitable
from abc import ABC, abstractmethod
from ..types import Update


class BaseMiddleware(ABC):
    def __init__(self, priority: int = 0):
        self.priority = priority

    @abstractmethod
    async def __call__(
        self,
        handler: Callable[[Update, Dict[str, Any]], Awaitable[Any]],
        update: Update,
        data: Dict[str, Any],
    ) -> Any:
        pass



import re
from typing import Union, Pattern
from ._base import _BaseTextFilter


class Regex(_BaseTextFilter):
    """Filter text/caption matching regex pattern"""

    def __init__(self, pattern: Union[str, Pattern], context: bool = False):
        compiled = re.compile(pattern)
        super().__init__(lambda t: bool(compiled.search(t)), context)



from typing import Optional
from ._base import _BaseTextFilter


class Text(_BaseTextFilter):
    """Filter text/caption matching exactly"""

    def __init__(self, text: Optional[str] = None, context: bool = False):
        if text is None:
            super().__init__(lambda _: True, context)
        else:
            super().__init__(lambda t: t == text, context)



from typing import Optional, Union, List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..types import (
        User,
        Message,
        InlineKeyboardMarkup,
        BotCommand,
        ChatMemberStatus,
    )


class Bot:
    def __init__(self, token: str):
        self.token = token
        self._get_me = None

    async def is_join(
        self,
        chat_id: Union[str, int],
    ) -> bool:
        from ..types._chat import ChatMemberStatus

        try:
            me = await self.get_me()
            status = await self.get_chat_member(chat_id=chat_id, user_id=me.id)
            if status and status in [
                ChatMemberStatus.ADMINISTRATOR,
                ChatMemberStatus.CREATOR,
                ChatMemberStatus.MEMBER,
                ChatMemberStatus.RESTRICTED,
            ]:
                return True
        except Exception:
            return False

    async def is_admin(
        self,
        chat_id: Union[str, int],
    ) -> bool:
        from ..types._chat import ChatMemberStatus

        try:
            me = await self.get_me()
            status = await self.get_chat_member(chat_id=chat_id, user_id=me.id)
            if status and status in [
                ChatMemberStatus.ADMINISTRATOR,
                ChatMemberStatus.CREATOR,
            ]:
                return True
        except Exception:
            return False

    async def get_me(self) -> "User":
        from .methods._get_me import GetMe

        if not self._get_me:
            me = await GetMe(self.token).execute()
            self._get_me = me
        return self._get_me

    async def set_webhook(
        self,
        url: str,
        max_connections: int = 40,
        allowed_updates: Optional[List[str]] = None,
        drop_pending_updates: bool = False,
        secret_token: Optional[str] = None,
    ) -> bool:
        from .methods._set_webhook import SetWebhook

        return await SetWebhook(self.token).execute(
            url=url,
            max_connections=max_connections,
            allowed_updates=allowed_updates,
            drop_pending_updates=drop_pending_updates,
            secret_token=secret_token,
        )

    async def restrict_user(
        self,
        chat_id: Union[int, str],
        user_id: int,
        until_date: int,
    ) -> bool:
        from .methods._restrict_user import RestrictUser

        return await RestrictUser(self.token).execute(
            chat_id=chat_id,
            user_id=user_id,
            until_date=until_date,
        )

    async def get_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
    ) -> "ChatMemberStatus":
        from .methods._get_chat_member import GetChatMember

        return await GetChatMember(self.token).execute(
            chat_id=chat_id,
            user_id=user_id,
        )

    async def set_my_commands(
        self,
        commands: list["BotCommand"],
    ) -> bool:
        from .methods._set_my_commands import SetMyCommands

        return await SetMyCommands(self.token).execute(
            commands=commands,
        )

    async def delete_webhook(
        self,
        drop_pending_updates: bool = False,
    ) -> bool:
        from .methods._delete_webhook import DeleteWebhook

        return await DeleteWebhook(self.token).execute(
            drop_pending_updates=drop_pending_updates,
        )

    async def send_message(
        self,
        chat_id: Union[int, str],
        text: str,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
        reply_to_message_id: Optional[int] = None,
    ) -> "Message":
        from .methods._send_message import SendMessage

        return await SendMessage(self.token).execute(
            chat_id=chat_id,
            text=text,
            reply_markup=reply_markup,
            reply_to_message_id=reply_to_message_id,
        )

    async def edit_message(
        self,
        chat_id: Union[int, str],
        message_id: int,
        text: str,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
    ) -> "Message":
        from .methods._edit_message import EditMessage

        return await EditMessage(self.token).execute(
            chat_id=chat_id, message_id=message_id, text=text, reply_markup=reply_markup
        )

    async def send_photo(
        self,
        chat_id: Union[int, str],
        photo: Union[str, bytes],
        caption: Optional[str] = None,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
    ) -> "Message":
        from .methods._send_photo import SendPhoto

        return await SendPhoto(self.token).execute(
            chat_id=chat_id, photo=photo, caption=caption, reply_markup=reply_markup
        )

    async def pin_message(
        self,
        chat_id: Union[int, str],
        message_id: int,
        disable_notification: bool = False,
    ) -> bool:
        from .methods._pin_message import PinMessage

        return await PinMessage(self.token).execute(
            chat_id=chat_id,
            message_id=message_id,
            disable_notification=disable_notification,
        )

    async def delete_messages(
        self, chat_id: Union[int, str], message_ids: List[int]
    ) -> List[bool]:
        from .methods._delete_messages import DeleteMessages

        return await DeleteMessages(self.token).execute(
            chat_id=chat_id, message_ids=message_ids
        )

    async def answer_callback(
        self,
        callback_query_id: str,
        text: Optional[str] = None,
        show_alert: Optional[bool] = None,
    ) -> bool:
        from .methods._answer_callback import AnswerCallbackQuery

        return await AnswerCallbackQuery(self.token).execute(
            callback_query_id=callback_query_id, text=text, show_alert=show_alert
        )



from enum import StrEnum
from typing import Optional, Union, List
from pydantic import BaseModel, Field
from ._user import User
from ._chat import Chat
from ._inline_keyboard import InlineKeyboardMarkup
from ..client import Bot


class EntityType(StrEnum):
    MENTION = "mention"
    HASHTAG = "hashtag"
    CASHTAG = "cashtag"
    BOT_COMMAND = "bot_command"
    URL = "url"
    EMAIL = "email"
    PHONE_NUMBER = "phone_number"
    BOLD = "bold"
    ITALIC = "italic"
    UNDERLINE = "underline"
    STRIKETHROUGH = "strikethrough"
    SPOILER = "spoiler"
    CODE = "code"
    PRE = "pre"
    TEXT_LINK = "text_link"
    TEXT_MENTION = "text_mention"


class PhotoSize(BaseModel):
    file_id: str
    file_unique_id: str
    width: int
    height: int
    file_size: Optional[int] = None


class MessageEntity(BaseModel):
    type: EntityType
    offset: int
    length: int
    url: Optional[str] = None
    user: Optional["User"] = None


class Message(BaseModel):
    message_id: int
    from_user: User = Field(..., alias="from")
    chat: Chat
    text: Optional[str] = None
    photo: Optional[list[PhotoSize]] = None
    caption: Optional[str] = None
    caption_entities: Optional[List[MessageEntity]] = None
    bot: Optional[Bot] = None
    entities: Optional[List[MessageEntity]] = None

    class Config:
        validate_by_name = True
        arbitrary_types_allowed = True

    @property
    def id(self) -> int:
        return self.message_id

    @property
    def context(self) -> Optional[str]:
        return self.text or self.caption

    def __str__(self) -> str:
        media_info = f", media={self.photo[0].file_id}" if self.photo else ""
        return f"Message(id={self.id}, text={self.context or 'None'}{media_info})"

    def html_text(self) -> str:
        from ..utils.html_parse import MessageParser

        return MessageParser.parse_to_html(
            text=self.context or "",
            entities=self.entities or self.caption_entities,
        )

    async def answer(
        self,
        text: str,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
    ) -> "Message":
        return await self.bot.send_message(
            chat_id=self.chat.id,
            text=text,
            reply_markup=reply_markup,
        )

    async def answer_photo(
        self,
        photo: Union[str, bytes],
        caption: Optional[str] = None,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
    ) -> "Message":
        return await self.bot.send_photo(
            chat_id=self.chat.id,
            photo=photo,
            caption=caption,
            reply_markup=reply_markup,
        )

    async def reply(
        self,
        text: str,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
    ) -> "Message":
        return await self.bot.send_message(
            chat_id=self.chat.id,
            text=text,
            reply_markup=reply_markup,
            reply_to_message_id=self.message_id,
        )

    async def edit(
        self,
        text: str,
        reply_markup: Optional["InlineKeyboardMarkup"] = None,
    ) -> "Message":
        return await self.bot.edit_message(
            chat_id=self.chat.id,
            message_id=self.message_id,
            text=text,
            reply_markup=reply_markup,
        )

    async def delete(self) -> bool:
        return await self.bot.delete_messages(
            chat_id=self.chat.id, message_ids=[self.message_id]
        )

    async def pin(self, disable_notification: bool = False) -> bool:
        return await self.bot.pin_message(
            chat_id=self.chat.id,
            message_id=self.message_id,
            disable_notification=disable_notification,
        )

    async def mute(self, until_date: int) -> bool:
        if self.from_user is None:
            return None
        return await self.bot.restrict_user(
            chat_id=self.chat.id,
            user_id=self.from_user.id,
            until_date=until_date,
        )



from typing import Union, Optional
from eiogram.types import Message, InlineKeyboardMarkup
from ._base import MethodBase


class SendPhoto(MethodBase):
    async def execute(
        self,
        chat_id: Union[int, str],
        photo: Union[str, bytes],
        caption: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None,
    ) -> Message:
        data = {
            "chat_id": chat_id,
            "photo": photo,
            "parse_mode": "HTML",
        }

        if caption:
            data["caption"] = caption
        if reply_markup:
            data["reply_markup"] = reply_markup.dict()

        response = await self._make_request("POST", "sendPhoto", data)
        return Message(**response["result"])
